#include <QList>
#include <cmath>
#include <algorithm>
#include <QtWidgets/QApplication>
#include <QtWidgets/QMainWindow>
#include <QtWidgets/QWidget>
#include <QtWidgets/QLabel>
#include <QtWidgets/QLineEdit>
#include <QtWidgets/QPushButton>
#include <QtWidgets/QCheckBox>
#include <QtWidgets/QTreeWidget>
#include <QtWidgets/QTreeWidgetItem>
#include <QtWidgets/QComboBox>
#include <QStandardItemModel>
#include <QStandardItem>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QFileDialog>
#include <QFont>
#include <QString>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonArray>
#include <QJsonValue>
#include <QFile>
#include <QMessageBox>
#include <QJsonParseError>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QUrl>
#include <QHeaderView>
#include <QScreen>
#include <QElapsedTimer>

// For timing functions
template<typename Func>
auto time_action(Func func) {
    QElapsedTimer timer;
    timer.start();
    func();
    return timer.elapsed();
}

using namespace std;
QString dataURL = "http://pvc-utils.xyz/resources/data.json";
QString currencyMultipliersURL = "http://pvc-utils.xyz/resources/currencymultipliers.json";
QString disclaimer = "This program is NOT AFFILIATED WITH OR MADE BY THE OWNERS OF PEACEFUL VANILLA CLUB.\n it is provided as is, without any warranty.\n If you have any questions, please contact me on discord: ' .ilovecanada '\n\n";

// Data struct to store trades
struct Trade {
    QString currency;
    QString currency2;
    int price_without_conversion;
    int price_without_conversion_2;
    double price;
    QString trade;
    int stock;
    int tradeAmount;
    QString name;
    QString location;
    QString shopOwner;
    int trade_id;
    int shop_id;
    bool shulker_box;
    QHash<QString, int> enchants;

    Trade(const QString& currency, const QString& currency2, int price_without_conversion, int price_without_conversion_2, double price, const QString& trade, int stock, int tradeAmount, const QString& name, const QString& location, const QString& shopOwner, int trade_id, int shop_id, bool shulker_box, const QHash<QString, int>& enchants) :
        currency(currency),
        currency2(currency2),
        price_without_conversion(price_without_conversion),
        price_without_conversion_2(price_without_conversion_2),
        price(price),
        trade(trade),
        stock(stock),
        tradeAmount(tradeAmount),
        name(name),
        location(location),
        shopOwner(shopOwner),
        trade_id(trade_id),
        shop_id(shop_id),
        shulker_box(shulker_box),
        enchants(enchants)
    {
    }
};


// Declaring some variables
QTreeWidget* tradesTreeview; // Treeview to display trades
QTreeWidget* resultTreeview ; // Treeview to display items
QList<Trade> trades;   // Stores the trades
QHash<QString, QList<double>> tradePrices; // Stores the prices per item
QString romanNumerals[] = {"", " I", " II", " III", " IV", " V", " VI", " VII", " VIII"};
QDialog* currentWindow = nullptr;


QJsonDocument downloadJson(QUrl URL){
    QNetworkAccessManager* manager = new QNetworkAccessManager();


    QNetworkRequest networkRequest(URL);
    QNetworkReply* reply = manager->get(networkRequest);

    QEventLoop eventLoop;
    QObject::connect(reply, &QNetworkReply::finished, &eventLoop, &QEventLoop::quit);
    eventLoop.exec();

    QByteArray replyData = reply->readAll();
    QJsonDocument jsonDocument = QJsonDocument::fromJson(replyData);
    reply->deleteLater();
    return jsonDocument;
}

double getMedian(const QList<double>& data) {
    if (data.empty()) {
        return 0.0; // Handle the case of an empty QList
    }

    QList<double> sortedData = data; // Create a copy of the data
    sort(sortedData.begin(), sortedData.end());

    size_t size = sortedData.size();
    if (size % 2 == 0) {
        // If the size is even, return the average of the two middle values
        size_t mid = size / 2;
        return (sortedData[mid - 1] + sortedData[mid]) / 2.0;
    } else {
        // If the size is odd, return the middle value
        return sortedData[size / 2];
    }
}

// Function to calculate the mean (average) of a QList of doubles
double getMean(const QList<double>& data) {
    if (data.empty()) {
        return 0.0; // Handle the case of an empty QList
    }

    double sum = 0.0;
    for (const double& value : data) {
        sum += value;
    }

    return sum / static_cast<double>(data.size());
}

// Function to calculate the standard deviation of a QList of doubles
double getStdDev(const QList<double>& data) {
    if (data.empty()) {
        return 0.0; // Handle the case of an empty QList
    }

    double mean = getMean(data);
    double sumSquaredDiffs = 0.0;

    for (const double& value : data) {
        double diff = value - mean;
        sumSquaredDiffs += diff * diff;
    }

    double variance = sumSquaredDiffs / static_cast<double>(data.size());
    return sqrt(variance);
}

// Normalize prices to emeralds
double normalizePrices(const QString& currency, double price, QHash<QString, double> currencyMultipliers) {

    auto it = currencyMultipliers.find(currency);
    if (it != currencyMultipliers.end()) {
        double multiplier = it.value();
        price *= multiplier;
    } else {
        price = 0;
    }
    return price;
}

uint hashTrade(int a, const QHash<QString, int>& hash) {
    uint seed = qHash(a);
    for (auto item = hash.begin(); item != hash.end(); ++item) {
        seed ^= qHash(item.key(), item.value());
    }
    return seed;
}

// Finds optimal display price and adjusts prices
// Avoids displaying prices such as "0.0001 emerald blocks per firework"
tuple<double, double, QString> adjustDisplayedPrice(double price, double price2, QHash<QString, double> currencyMultipliers) {
    QString currency;
    if (price > currencyMultipliers["DIAMOND_BLOCK"]) {
        currency = "Diamond Blocks";
        price /= currencyMultipliers["DIAMOND_BLOCK"];
        price2 /= currencyMultipliers["DIAMOND_BLOCK"];
    } else if (price > currencyMultipliers["DIAMOND"]) {
        currency = "Diamonds";
        price /= currencyMultipliers["DIAMOND"];
        price2 /= currencyMultipliers["DIAMOND"];
    } else if (price < currencyMultipliers["EMERALD_BLOCK"]) {
        currency = "Emerald Blocks / Stack";
        price *= 64;
        price2 *= 64;
    } else {
        currency = "Emerald Blocks";
    }

    return make_tuple(price, price2, currency);
}



// Main parsing logic starts here
// Parses the JSON data, extracts all trades and values, populates a list with trade instances,
// Populates a map with items and lists of prices
void parse(const QJsonObject& data, QList<Trade>& trades, QHash<QString, QList<double>>& tradePrices, QHash<QString, double> currencyMultipliers, QList<QString> validCurrencies) {
    qDebug() << "Parsing JSON data...";
    auto startTime = chrono::high_resolution_clock::now();

    trades.clear();
    tradePrices.clear();

    int trade_id = 0;
    int shop_id = 0;

    if (data.contains("data") && data["data"].isArray()) {
        QJsonArray dataArray = data["data"].toArray();

        for (const QJsonValue& entryValue : dataArray) {
            QJsonObject entry = entryValue.toObject();

            // For each trade fetch all necessary values
            if (entry.contains("recipes") && entry["recipes"].isArray()) {
                QJsonArray recipesArray = entry["recipes"].toArray();
                for (const QJsonValue& recipeValue : recipesArray) {
                    QJsonObject recipe = recipeValue.toObject();
                    bool shulker_box = false;

                    QJsonObject item1 = recipe["item1"].toObject();
                    QString currency = item1["type"].toString();
                    double price = item1["amount"].toDouble();
                    QString currency2;
                    double price2 = 0.0;

                    if (recipe.contains("item2")) {
                        QJsonObject item2 = recipe["item2"].toObject();
                        currency2 = item2["type"].toString();
                        price2 = item2["amount"].toDouble();
                    }

                    QJsonObject resultItem = recipe["resultItem"].toObject();
                    QString trade = resultItem["type"].toString();
                    int tradeAmount = resultItem["amount"].toInt();

                    // Make a map with enchantments
                    QJsonObject enchantObject = resultItem["enchant"].toObject();
                    QHash<QString, int> enchants;
                    for (auto it = enchantObject.begin(); it != enchantObject.end(); ++it) {
                        enchants.insert(it.key(), it.value().toInt());
                    }
                    int stock = recipe["stock"].toInt() * tradeAmount;
                    QString name = entry["shopName"].toString();
                    QString location = entry["location"].toString();
                    QString world = entry["world"].toString();
                    QString shopOwner = entry["shopOwner"].toString();

                    // Format the location
                    if (world == "World") {
                        location += " in Overworld";
                    } else if (world == "World_nether") {
                        location += " in Nether";
                    } else {
                        location += " in End";
                    }

                    bool validity = true; // Flag to track if the trade is valid

                    // Handle shulker boxes
                    // Adds all items inside the shulker box and sets the trade item
                    // To shulker box contents
                    // Ignores shulkers with several item types inside
                    if (trade.contains("SHULKER_BOX")) {
                        shulker_box = true;
                        QJsonArray loreEntries = recipe["resultItem"].toObject()["lore"].toArray();
                        if (!loreEntries.isEmpty()) {
                            // Check if all lore entries are identical
                            bool identical = adjacent_find(loreEntries.begin(), loreEntries.end(), not_equal_to<>()) == loreEntries.end();
                            if (identical) {

                                // Reinitialize stock and trade
                                stock = 0;
                                trade = "";
                                tradeAmount = 0;

                                // Extract the name and amount from the lore entry
                                QString loreEntry = loreEntries[0].toString().mid(2); // Remove the leading '- '
                                QStringList parts = loreEntry.split(" ", Qt::SkipEmptyParts);

                                if (parts.size() >= 2) {
                                    try {
                                        int lore_entry_multiplier = parts[0].mid(0, parts[0].length() - 1).toInt(); // Extract the lore entry multiplier by removing the 'x' character
                                        QString lore_text = parts[1];

                                        // Append item name in the lore entry to trade
                                        tradeAmount += lore_entry_multiplier * loreEntries.size();
                                        stock = lore_entry_multiplier * loreEntries.size();
                                        trade = lore_text;
                                    } catch (const invalid_argument& e) {
                                        qDebug() << "Invalid lore entry format: " << loreEntry;
                                        validity = false; // Set the flag to indicate an invalid trade
                                    }
                                } else {
                                    validity = false;
                                }
                                stock *= recipe["stock"].toInt();
                            } else {
                                validity = false;
                            }
                        }
                    }

                    double normalizedPrice = price;
                    double normalizedPrice2 = price2;
                    // Handle invalid currencies
                    if (!validCurrencies.contains(currency)) {
                        normalizedPrice = 0.0;

                    }
                    if (!currency2.isEmpty() && !validCurrencies.contains(currency2)) {
                        normalizedPrice2 = 0.0;

                    }
                    normalizedPrice = normalizePrices(currency, normalizedPrice, currencyMultipliers) + normalizePrices(currency2, normalizedPrice2, currencyMultipliers);
                    if (normalizedPrice == 0.0) {
                        validity = false;
                    }

                    double pricePerUnit = normalizedPrice / tradeAmount;

                    // Create the trade instance and add price to prices dictionary
                    if (validity) {

                        Trade tradeInstance(
                            currency,
                            currency2,
                            price,
                            price2,
                            pricePerUnit,
                            trade,
                            stock,
                            tradeAmount,
                            name,
                            location,
                            shopOwner,
                            trade_id,
                            shop_id,
                            shulker_box,
                            enchants
                        );

                        trade_id++;
                        trades.emplace_back(tradeInstance);
                        if (tradePrices.find(trade) == tradePrices.end()) {
                            tradePrices[trade] = QList<double>();
                        }

                        tradePrices[trade].emplace_back(pricePerUnit);
                    }
                }
                shop_id++;
            }
        }
    }

    QList<Trade> tradesToDelete;
    int price_threshold = 10;
    // For each item in the tradePrices dictionary
    for (auto item = tradePrices.begin(); item != tradePrices.end(); ++item) {
        const QString& tradeItem = item.key();
        const QList<double>& prices = item.value();

        // Calculate the median price§§
        double medianPrice = getMedian(prices);

        // For each trade in the trade list
        for (auto it = trades.begin(); it != trades.end(); ) {
            Trade& trade = *it;

            // If the trade matches the trade item
            if (trade.trade == tradeItem) {
                // If trade price is too high
                if (trade.price > price_threshold * medianPrice) {
                    tradesToDelete.emplace_back(trade);

                    // Remove the trade
                    it = trades.erase(it);
                    continue;
                }
            }
            ++it;
        }
    }

    qDebug() << "Deleted " << tradesToDelete.size() << " trades out of " << trades.size() + tradesToDelete.size() << " total trades because of a disproportionate price, most likely donations";

    qDebug() << "Total trades: " << trades.size();
    auto endTime = chrono::high_resolution_clock::now();
    auto duration = chrono::duration_cast<chrono::milliseconds>(endTime - startTime);
    qDebug() << "Finished parsing JSON data in " << duration;
}

// END OF PARSING SECTION
// BEGINNING OF GUI SECTION


// Show trades in the bottom treeview
void showTrades(QTreeWidgetItem* item, bool includeZeroStock, const QList<Trade>& trades, QTreeWidget* tradesTreeview, QHash<QString, double> currencyMultipliers) {
    if (!item) {
        return;
    }

    QString selected_item = item->text(0);
    tradesTreeview->clear();

    for (const Trade& trade : trades) {
        if (((!includeZeroStock && trade.stock != 0) || includeZeroStock) && trade.trade == selected_item) {
            QString tradeText = trade.trade;
            QString nameText = trade.name;
            QString locationText = trade.location;

            tuple priceInfo = adjustDisplayedPrice(trade.price, trade.price, currencyMultipliers);
            auto [displayPrice, _, currency] = priceInfo;

            QTreeWidgetItem* itemRow = new QTreeWidgetItem(tradesTreeview);
            itemRow->setText(0, tradeText);
            itemRow->setText(1, nameText);
            itemRow->setText(2, locationText);
            itemRow->setData(3, Qt::DisplayRole, round(displayPrice * 100) / 100);
            itemRow->setText(4, currency);
            itemRow->setData(5, Qt::DisplayRole, trade.stock);
            // Hidden columns 6 and 7 for trade ID and unadjusted trade price
            itemRow->setData(6, Qt::DisplayRole, trade.trade_id);
            itemRow->setData(7, Qt::DisplayRole, trade.price);
        }
    }
}

// Search for items and display them in the top treeview
void searchItem(const QString& inputText, bool includeZeroStock, QTreeWidget* resultTreeview, const QHash<QString, QList<double>>& tradePrices, QHash <QString, double> currencyMultipliers) {
    QString item = inputText.toUpper().replace(" ", "_");
    resultTreeview->clear();

    for (auto itemPrice = tradePrices.begin(); itemPrice != tradePrices.end(); ++itemPrice) {
        if (itemPrice.key().contains(item)) {
            // If the first value of itemPrice (aka item as itemPrice is an item: price pair)
            // Matches with the input string
            // Add it to the list of matching items
            QList<double> filteredTradePrices;
            QList<double> tradeCurrencyMultipliers;

            // We do not want to add an item twice to stock if a shop is listed twice, so we keep track of already listed shop IDs, item enchants
            // And already listed shop stocks, all hashed into a hashmap with shop ID as key
            QHash<int, QSet<uint>> tradeHashes;

            int totalStock = 0;

            for (const Trade& trade : trades) {
                if (trade.trade == itemPrice.key()) {
                    if (!includeZeroStock && (trade.stock == 0)) {
                        continue;
                    }
                    filteredTradePrices.emplace_back(trade.price);
                    tradeCurrencyMultipliers.emplace_back(currencyMultipliers[trade.currency] / trade.tradeAmount);

                    bool unique = false;
                    uint tradeHash = hashTrade(trade.stock, trade.enchants);

                    // if key exists and value contains tradeHash, trade is duplicate
                    auto it = tradeHashes.find(trade.shop_id);
                    if (it != tradeHashes.end() && it.value().contains(tradeHash)) {
                        continue;
                    } else {
                        unique = true;
                        tradeHashes[trade.shop_id].insert(tradeHash);
                    }

                    if (unique) {
                        totalStock += trade.stock;
                    }
                }
            }

            if (!filteredTradePrices.empty()) {
                double medianPrice = getMedian(filteredTradePrices);
                double meanPrice = getMean(filteredTradePrices);
                double avgCurrencyMultiplier = getMean(tradeCurrencyMultipliers);

                double unadjustedMeanPrice = meanPrice;
                double unadjustedMedianPrice = medianPrice;

                double stdDevPrice = getStdDev(filteredTradePrices);
                tuple priceInfo = adjustDisplayedPrice(medianPrice, meanPrice, currencyMultipliers);
                auto [adjustedMedianPrice, adjustedMeanPrice, currency] = priceInfo;
                // Populate the treeview
                QTreeWidgetItem* itemRow = new QTreeWidgetItem(resultTreeview);
                itemRow->setText(0, itemPrice.key());
                itemRow->setData(1, Qt::DisplayRole, round(adjustedMedianPrice * 100) / 100);
                itemRow->setData(2, Qt::DisplayRole, round(adjustedMeanPrice * 100) / 100);
                itemRow->setText(3, currency);
                itemRow->setData(4, Qt::DisplayRole, totalStock);
                if (stdDevPrice > 0) {
                    itemRow->setData(5, Qt::DisplayRole, round((stdDevPrice / avgCurrencyMultiplier) * 100) / 100);
                }
                // Hidden columns 6 and 7 for unadjusted price
                itemRow->setData(6, Qt::DisplayRole, unadjustedMedianPrice);
                itemRow->setData(7, Qt::DisplayRole, unadjustedMeanPrice);
            }
        }
    }
}


void showTradeDetails(QTreeWidgetItem* item, QList<Trade>& trades, QMainWindow* mainWindow, QHash<QString, double> currencyMultipliers) {
    if (!item) {
        return;
    }

    int selectedTrade_id = item->text(6).toInt();
    // Find the selected trade by trade ID
    Trade* selectedTrade = nullptr;
    for (Trade& trade : trades) {
        if (trade.trade_id == selectedTrade_id) {
            selectedTrade = &trade;
            break;
        }
    }

    if (selectedTrade) {

        // Call the adjustDisplayedPrice function to get the adjusted price and currency
        tuple priceInfo = adjustDisplayedPrice(selectedTrade->price, selectedTrade->price, currencyMultipliers);
        auto [displayPrice, _, currency] = priceInfo;

        // Close any already existing popup window
        if (currentWindow) {
            currentWindow->close();
            currentWindow->deleteLater();
        }

        // Create a window to display trade details
        QDialog* tradeDetailsWindow = new QDialog(mainWindow);
        tradeDetailsWindow->setModal(false); // Prevent the window from blocking input to the main window
        tradeDetailsWindow->setWindowTitle("Trade Details");
        currentWindow = tradeDetailsWindow;

        // Create a new Treeview widget
        QTreeWidget* table = new QTreeWidget(tradeDetailsWindow);
        table->setColumnCount(2);

        QScreen* screen = QGuiApplication::primaryScreen();
        int screenHeight = screen->size().height();
        int screenWidth = screen->size().width();


        table->setHeaderLabels(QStringList() << "Information" << "Value");

        // Handle enchantments
        QString enchantsString;
        for (auto it = selectedTrade->enchants.begin(); it != selectedTrade->enchants.end(); it++) {
            QString enchantName = it.key();
            int enchantLevel = it.value();

            // Remove the "minecraft:" prefix and underscores with spaces
            enchantName.remove(0, 10).replace("_", " ");

            // Replace numbers with Roman numerals
            enchantName += romanNumerals[enchantLevel];

            // Concatenenate the enchantment to the string
            if (!enchantsString.isEmpty()) {
                enchantsString += ", ";
            }
            enchantsString += enchantName;
        }

        // Prepare key: value pairs
        QList<QPair<QString, QString>> values;
        values.append(qMakePair(QString("Item"), QString("%1x %2").arg(selectedTrade->tradeAmount).arg(selectedTrade->trade)));
        values.append(qMakePair(QString("Price 1"), QString("%1x %2").arg(selectedTrade->price_without_conversion).arg(selectedTrade->currency)));
        if (!selectedTrade->currency2.isEmpty()) {
            values.append(qMakePair(QString("Price 2"), QString("%1x %2").arg(selectedTrade->price_without_conversion_2).arg(selectedTrade->currency2)));
        }
        values.append(qMakePair(QString("Adjusted Price per item"), QString::number(displayPrice, 'f', 2) + " " + currency));
        values.append(qMakePair(QString("Stock"), QString::number(selectedTrade->stock)));
        values.append(qMakePair(QString("Shop Owner"), selectedTrade->shopOwner));
        values.append(qMakePair(QString("Shop Name"), selectedTrade->name));
        values.append(qMakePair(QString("Location"), selectedTrade->location));
        values.append(qMakePair(QString("In shulker box"), selectedTrade->shulker_box ? "yes" : "no"));
        values.append(qMakePair(QString("Enchants"), enchantsString));

        // Add the pairs to the table
        for (const QPair<QString, QString>& pair : values) {
            if (!pair.second.isEmpty()){
                QTreeWidgetItem* item = new QTreeWidgetItem(table);
                item->setText(0, pair.first);
                item->setText(1, pair.second);
                table->addTopLevelItem(item);
            }
        }

        // Gather variables to create the map link
        QStringList positions = selectedTrade->location.split(" ");
        QStringList coordinates;
        for (int i = 0; i < 3; i++) {
            QString num = positions[i].replace(",", "");
            double roundedNum = round(num.toDouble());
            coordinates.append(QString::number(roundedNum));
        }
        QString world = positions[4];

        QString dimension;
        if (world == "Overworld") {
            dimension = "World";
        } else if (world == "Nether") {
            dimension = "World_nether";
        } else {
            dimension = "World_end";
        }

        QString map_link = QString("https://web.peacefulvanilla.club/maps/#%1;flat;%2;6").arg(dimension).arg(coordinates.join(","));

        // Add the map link to the table
        QTreeWidgetItem* mapLinkItem = new QTreeWidgetItem(table);
        mapLinkItem->setText(0, "Map Link");

        QLabel* mapLinkLabel = new QLabel();
        mapLinkLabel->setText(QString("<a href=\"%1\">%1</a>").arg(map_link));
        mapLinkLabel->setOpenExternalLinks(true);

        table->setItemWidget(mapLinkItem, 1, mapLinkLabel);

        // Set the layout for the dialog window
        QVBoxLayout* layout = new QVBoxLayout();
        layout->addWidget(table);
        tradeDetailsWindow->setLayout(layout);

        table->resizeColumnToContents(0);
        table->resizeColumnToContents(1);

        if(screenWidth > screenHeight * 1.33) {
            tradeDetailsWindow->setGeometry(100, 100, screenWidth / 3.5, screenHeight / 4.5);
        } else {
            tradeDetailsWindow->setGeometry(100, 100, screenWidth / 1.5, screenHeight / 5);
        }
        // Show the dialog window
        tradeDetailsWindow->show();
    }
}

void searchTrades(const QString& searchedTrade, const QString& searchedShopName, const QString& searchedShopOwner, bool includeZeroStock, QTreeWidget* treeview, QList<Trade>& trades, QHash<QString, double> currencyMultipliers) {

    QString shopNameInput = searchedShopName.toLower().replace(" ","");
    QString shopOwnerInput = searchedShopOwner.toLower().replace("_", "");
    QString tradeInput = searchedTrade.toLower().replace("_"," ");

    treeview->clear();

    for (const Trade& trade : trades) {
        // Format the trade characteristics for searching
        QString shopName = trade.name.toLower().replace(" ","");
        QString shopOwner = trade.shopOwner.toLower().replace("_", "");
        QString tradeItem = trade.trade.toLower().replace("_"," ");

        // Filter trades based on the provided search strings
        bool shopNameMatch = shopNameInput.isEmpty() || shopName.contains(shopNameInput);
        bool shopOwnerMatch = shopOwnerInput.isEmpty() || shopOwner.contains(shopOwnerInput);
        bool tradeMatch = tradeInput.isEmpty() || tradeItem.contains(tradeInput);

        if (((!includeZeroStock && trade.stock != 0) || includeZeroStock) && shopNameMatch && shopOwnerMatch && tradeMatch) {
            // Adapt price and currency to the most suitable display currency
            tuple priceInfo = adjustDisplayedPrice(trade.price, trade.price, currencyMultipliers);
            auto [displayPrice, _, currency] = priceInfo;

            // Populate the treeview
            QTreeWidgetItem* itemRow = new QTreeWidgetItem(treeview);
            itemRow->setText(0, trade.trade);
            itemRow->setText(1, trade.name);
            itemRow->setText(2, trade.location);
            itemRow->setData(3, Qt::DisplayRole, round(displayPrice*100)/100);
            itemRow->setText(4, currency);
            itemRow->setData(5, Qt::DisplayRole, trade.stock);

            itemRow->setText(6, QString::number(trade.trade_id));
            itemRow->setText(7, QString::number(trade.price));
        }
    }
}



int main(int argc, char* argv[]) {
    QApplication app(argc, argv);
    qDebug() << disclaimer;

    int columnCount = 6;
    bool ascending = 1;

    // Fetch currency multipliers
    QJsonObject currencyMultipliersJson = downloadJson(currencyMultipliersURL).object();
    QHash<QString, double> currencyMultipliers;
    QList<QString> validCurrencies;

    for (auto it = currencyMultipliersJson.begin(); it != currencyMultipliersJson.end(); ++it) {
        QString key = it.key();
        QJsonValue value = it.value();

        // Ensure that the value is a double before inserting
        if (value.isDouble()) {
            currencyMultipliers.insert(key, value.toDouble());
            validCurrencies.emplace_back(key);
        }
    }
    parse(downloadJson(dataURL).object(), trades, tradePrices, currencyMultipliers, validCurrencies);


    // Set application color palette
    QPalette darkPalette;
    darkPalette.setColor(QPalette::Window, QColor(9, 4, 11));
    darkPalette.setColor(QPalette::WindowText, Qt::white);
    darkPalette.setColor(QPalette::Base, QColor(18, 12, 22));
    darkPalette.setColor(QPalette::AlternateBase, QColor(20, 18, 20));

    darkPalette.setColor(QPalette::ToolTipBase, Qt::white);
    darkPalette.setColor(QPalette::ToolTipText, Qt::white);
    darkPalette.setColor(QPalette::Text, Qt::white);
    darkPalette.setColor(QPalette::PlaceholderText, Qt::lightGray);
    darkPalette.setColor(QPalette::Button, QColor(25, 21, 32));
    darkPalette.setColor(QPalette::ButtonText, Qt::white);
    darkPalette.setColor(QPalette::BrightText, Qt::red);
    darkPalette.setColor(QPalette::Link, QColor(75, 75, 255));

    darkPalette.setColor(QPalette::Highlight, Qt::lightGray);
    darkPalette.setColor(QPalette::HighlightedText, QColor(15, 6, 10));

    qApp->setPalette(darkPalette);

    QScreen* screen = QGuiApplication::primaryScreen();

    // Get screen width and height
    int screenWidth = screen->size().width();
    int screenHeight = screen->size().height();

    // Set the font size based on the screen DPI
    qreal dpi = screen->logicalDotsPerInch();
    int fontSize = dpi / 96 * 9; // Scale font
    QFont globalFont;
    globalFont.setPointSize(fontSize);
    app.setFont(globalFont);
    app.setStyleSheet("QLineEdit { border: 1px solid rgb(75, 75, 75); height: " + QString::number(screenHeight/35)+ "px;} QQPushButton, QCheckBox { border: 1px solid rgb(75, 75, 75); }");


    // Create the trades treeview
    tradesTreeview = new QTreeWidget();
    tradesTreeview->setColumnCount(columnCount + 2);
    tradesTreeview->setHeaderLabels(QStringList() << "Item" << "Shop Name" << "Location" << "Price" << "Normalized Currency" << "Stock");
    tradesTreeview->setSortingEnabled(true);
    tradesTreeview->setColumnHidden(6, true);
    tradesTreeview->setColumnHidden(7, true);

    // Create the main window
    QMainWindow* mainWindow = new QMainWindow();
    QWidget* centralWidget = new QWidget();
    QVBoxLayout* mainLayout = new QVBoxLayout();

    // Create the input fields
    QLineEdit* entry = new QLineEdit();

    QLineEdit* tradeSearchEntry = new QLineEdit();
    QLineEdit* shopOwnerEntry = new QLineEdit();
    QLineEdit* shopNameEntry = new QLineEdit();

    entry->setPlaceholderText("Search for an item...");
    tradeSearchEntry->setPlaceholderText("Search for a trade...");
    shopOwnerEntry->setPlaceholderText("Search for a shop owner...");
    shopNameEntry->setPlaceholderText("Search for a shop name...");

    QCheckBox* includeZeroStockCheckBox = new QCheckBox("Include out-of-stock trades");

    QHBoxLayout* bottomBarLayout = new QHBoxLayout();
    QHBoxLayout* middleBarLayout = new QHBoxLayout();

    bottomBarLayout->addWidget(includeZeroStockCheckBox);

    // Create the result treeview
    resultTreeview = new QTreeWidget();
    resultTreeview->setColumnCount(columnCount + 2);
    resultTreeview->setHeaderLabels(QStringList() << "Item" << "Median Price" << "Average Price" << "Currency" << "Total Stock" << "Price Std. Dev.");
    resultTreeview->setSortingEnabled(true);
    resultTreeview->setColumnHidden(6, true);
    resultTreeview->setColumnHidden(7, true);

    // Add widgets to layouts and layouts to main window
    mainLayout->addWidget(entry);
    middleBarLayout->addWidget(tradeSearchEntry);
    middleBarLayout->addWidget(shopNameEntry);
    middleBarLayout->addWidget(shopOwnerEntry);
    mainLayout->addWidget(resultTreeview);

    mainLayout->addLayout(middleBarLayout);
    mainLayout->addWidget(tradesTreeview);
    mainLayout->addLayout(bottomBarLayout);

    // Store what column to sort
    int currentSortColumn = 0;

    // Handle events
    QObject::connect(entry, &QLineEdit::textChanged, [&]() {
        qDebug() << "Action took " << time_action([&](){
            searchItem(entry->text(), includeZeroStockCheckBox->isChecked(), resultTreeview, tradePrices, currencyMultipliers);
        }) << "ms";
    });

    auto tradeSearchAction = [&]() {
        qDebug() << "Action took " << time_action([&](){
            searchTrades(tradeSearchEntry->text(), shopNameEntry->text(), shopOwnerEntry->text(), includeZeroStockCheckBox->isChecked(), tradesTreeview, trades, currencyMultipliers);
        }) << "ms";
    };

    QObject::connect(tradeSearchEntry, &QLineEdit::textChanged, tradeSearchAction);
    QObject::connect(shopOwnerEntry, &QLineEdit::textChanged, tradeSearchAction);
    QObject::connect(shopNameEntry, &QLineEdit::textChanged, tradeSearchAction);

    QObject::connect(includeZeroStockCheckBox, &QCheckBox::stateChanged, [&](int state) {
        qDebug() << "Action took " << time_action([&](){
            searchItem(entry->text(), state == Qt::Checked, resultTreeview, tradePrices, currencyMultipliers);
        }) << "ms";

        tradesTreeview->clear();
    });

    QObject::connect(tradesTreeview, &QTreeWidget::itemClicked, [&](QTreeWidgetItem* item) {
        qDebug() << "Action took " << time_action([&](){
            showTradeDetails(item, trades, mainWindow, currencyMultipliers);
        }) << "ms";
    });

    QObject::connect(resultTreeview, &QTreeWidget::itemClicked, [&](QTreeWidgetItem* item) {
        qDebug() << "Action took " << time_action([&](){
            showTrades(item, includeZeroStockCheckBox->isChecked(), trades, tradesTreeview, currencyMultipliers);
        }) << "ms";
    });

    QObject::connect(resultTreeview->header(), &QHeaderView::sectionClicked, [&](int logicalIndex) {
        currentSortColumn = logicalIndex;
        // Sort price columns by unadjusted price and not display price
        currentSortColumn = (currentSortColumn == 1) ? 6 : (currentSortColumn == 2) ? 7 : currentSortColumn;
        resultTreeview->sortItems(currentSortColumn, ascending ? Qt::AscendingOrder : Qt::DescendingOrder);
        ascending = !ascending;
    });

    QObject::connect(tradesTreeview->header(), &QHeaderView::sectionClicked, [&](int logicalIndex) {
        currentSortColumn = logicalIndex;
        if (currentSortColumn == 3){
            currentSortColumn = 7;
        }
        tradesTreeview->sortItems(currentSortColumn, ascending ? Qt::AscendingOrder : Qt::DescendingOrder);
        ascending = !ascending;
    });


    // Set the central widget and main window properties
    centralWidget->setLayout(mainLayout);
    mainWindow->setCentralWidget(centralWidget);
    mainWindow->setWindowTitle("Trade Search");
    mainWindow->show();

    // Dynamically resize columns and UI elements
    int columnWidth = screenWidth / columnCount;

    for (int i = 0; i < columnCount; i++) {
        tradesTreeview->setColumnWidth(i, (columnWidth > 125) ? columnWidth : 125);
        resultTreeview->setColumnWidth(i, (columnWidth > 125) ? columnWidth : 125);
    }

    return app.exec();
}
